﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace LinqToObjectsExample
{
    class Jednostka
    {
        public Jednostka() { K_Osoba = new List<Osoba>(); }
        public int Id { get; set; }
        public string Nazwa { get; set; }
        public List<Osoba> K_Osoba { get; set; }
    }
    class Osoba
    {
        public int Id { get; set; }
        public string Imie { get; set; }
        public string Nazwisko { get; set; }
        public Jednostka Jednostka { get; set; }
    }

    class Program
    {
        static void Main( string[] args )
        {
            Jednostka[] jednostki = new Jednostka[]
            {
                new Jednostka { Id = 1, Nazwa = "Warszawa" },
                new Jednostka { Id = 2, Nazwa = "Wrocław" },
                new Jednostka { Id = 3, Nazwa = "Kraków" },
            };
            Osoba[] osoby = new Osoba[]
            {
                new Osoba { Id = 1, Imie = "Jan01", Nazwisko = "Kowalski01", Jednostka = jednostki[0] },
                new Osoba { Id = 2, Imie = "Jan02", Nazwisko = "Kowalski02", Jednostka = jednostki[0] },
                new Osoba { Id = 3, Imie = "Jan03", Nazwisko = "Kowalski03", Jednostka = jednostki[1] },
                new Osoba { Id = 4, Imie = "Jan04", Nazwisko = "Kowalski04", Jednostka = jednostki[1] },
                new Osoba { Id = 5, Imie = "Jan05", Nazwisko = "Kowalski05", Jednostka = jednostki[1] }
            };
            jednostki[0].K_Osoba.Add( osoby[0] );
            jednostki[0].K_Osoba.Add( osoby[1] );
            jednostki[1].K_Osoba.Add( osoby[2] );
            jednostki[1].K_Osoba.Add( osoby[3] );
            jednostki[1].K_Osoba.Add( osoby[4] );

            // Konstrukcja

            // Restrykcja (Where)

            {
                int[] numbers = { 5, 4, 1, 3, 9, 8, 6, 7, 2, 0 };
                var res1 =
                    from n in numbers
                    where ( n < 5 ) && ( n > 2 )
                    select n;

                foreach ( int x in res1 )
                    Console.WriteLine( x );

                var res2 =
                    from o in osoby
                    where o.Jednostka == jednostki[0]
                    select o;

                foreach ( var o in res2 )
                    Console.WriteLine( o.Nazwisko );
            }

            // Projekcja (Select, SelectMany)

            {
                int[] numbers = { 2, 3, 4, 1, 5 }; // pierwszy element określa ilość
                string[] nazwy = { "jeden", "dwa", "trzy", "cztery", "pięć" };
                var res =
                    from n in numbers
                    select nazwy[n - 1];

                foreach ( var s in res )
                    Console.WriteLine( s );
            }

            {
                var res =
                    from o in osoby
                    select new { Numer = o.Id, Opis = string.Format( "{0} {1}", o.Imie, o.Nazwisko ) };
                foreach ( var s in res )
                    Console.WriteLine( s.Opis );
            }

            {
                IEnumerable<Osoba> res1 =
                    jednostki.
                    Where( j => j.Nazwa == "Warszawa" ).
                    SelectMany( j => j.K_Osoba );

                IEnumerable<List<Osoba>> res2 =
                    jednostki.
                    Where( j => j.Nazwa == "Warszawa" ).
                    Select( j => j.K_Osoba );

                IEnumerable<Osoba> res3 =
                    from j in jednostki
                    where j.Nazwa == "Warszawa"
                    from o in j.K_Osoba
                    select o;
            }

            {
                IEnumerable<string> jednostkiOsoby =
                        jednostki.SelectMany(
                            ( jednostka, jednostkaIndex ) =>
                            jednostka.K_Osoba.Select( o => "Jednostka #" + ( jednostkaIndex + 1 ) +
                                                    " ma osobę o numerze " + o.Id ) );

                foreach ( var s in jednostkiOsoby )
                    Console.WriteLine( s );
            }

            // Podział, stronicowanie ( Take, Skip, TakeWhile, SkipWhile )

            {
                var res = (
                    from o in osoby
                    select o
                ).Skip( 2 ).Take( 2 );
            }
            {
                int[] numbers = { 5, 4, 1, 3, 9, 8, 6, 7, 2, 0 };
                var res = (
                    from n in numbers
                    select n
                    )
                    .SkipWhile( n => n < 6 )
                    .TakeWhile( n => n >= 6 );
            }
            {
                int[] numbers = { 5, 4, 1, 3, 9, 8, 6, 7, 2, 0 };
                var laterNumbers = numbers.SkipWhile( ( n, index ) => n >= index );
                foreach ( var n in laterNumbers )
                    Console.WriteLine( n );
            }

            // Złaczenia ( Join, GroupJoin )
            {
                var res =
                    from j in jednostki
                    join o in osoby on j.Id equals o.Jednostka.Id
                    select o;
                foreach ( var x in res )
                    Console.WriteLine( x.Imie + " " + x.Nazwisko + " " + x.Jednostka.Nazwa );

            }

            {
                // Źródło: http://www.hookedonlinq.com/OuterJoinSample.ashx
                var q = from j in jednostki
                        join o in osoby on j.Id equals o.Jednostka.Id into g
                        from o in g.DefaultIfEmpty()
                        select new { Nazwa = j.Nazwa, Nazwisko = o == null ? "(brak osób)" : o.Nazwisko };

                foreach ( var i in q )
                {
                    Console.WriteLine( "Jednostka: {0}  Numer osoby: {1}",
                        i.Nazwa.PadRight( 11, ' ' ), i.Nazwisko );
                }

            }

            // Operator Concat (łączy tablice)
            {
                int[] numbersA = { 0, 2, 4, 5, 6, 8, 9 };
                int[] numbersB = { 1, 3, 5, 7, 8 };

                var allNumbers = numbersA.Concat( numbersB );

                foreach ( var n in allNumbers )
                    Console.WriteLine( n );
            }

            // Porządkowanie ( OrderBy, ThenBy, Reverse )
            {
                var res =
                    from o in osoby
                    orderby o.Nazwisko descending, o.Jednostka.Nazwa ascending
                    select o;

            }

            {
                string[] digits = { "zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine" };

                var sortedDigits =
                    from d in digits
                    orderby d.Length, d
                    select d;

                foreach ( var d in sortedDigits )
                    Console.WriteLine( d );

                foreach ( var d in digits.Reverse() )
                    Console.WriteLine( d );
            }

            // Grupowanie ( GroupBy )

            {
                int[] numbers = { 5, 4, 1, 3, 9, 8, 6, 7, 2, 0 };

                var numberGroups =
                    from n in numbers
                    group n by n % 5 into g
                    select new { Remainder = g.Key, Numbers = g };

                foreach ( var g in numberGroups )
                {
                    Console.WriteLine( "Liczby z resztą {0} przy podziale przez 5:", g.Remainder );
                    foreach ( var n in g.Numbers )
                        Console.WriteLine( n );
                }
            }

            {
                string[] words = { "blueberry", "chimpanzee", "abacus", "banana", "apple", "cheese" };

                var wordGroups =
                    from w in words.OrderBy( s => s )
                    group w by w[0] into g
                    select new { FirstLetter = g.Key, Words = g };

                foreach ( var g in wordGroups )
                {
                    Console.WriteLine( "Słowa zaczynające się na literę '{0}':", g.FirstLetter );
                    foreach ( var w in g.Words )
                        Console.WriteLine( w );
                }
            }

            {
                string[] anagrams = { "from ", "salt", "earn ", " last ", " near ", " form " };

                var orderGroups = anagrams.GroupBy(
                    w => w.Trim(),
                    a => a.ToUpper(),
                    new AnagramEqualityComparer() );
                foreach ( var g in orderGroups )
                {
                    Console.WriteLine( "--" );
                    foreach ( var s in g )
                        Console.WriteLine( s );
                }
            }

            // Operatory zbiorów ( Distinct, Union, Intersect, Except )
            {
                int[] factors = { 2, 2, 3, 5, 5 };
                var uniqueFactors = factors.Distinct();
                foreach ( var f in uniqueFactors )
                    Console.WriteLine( f );
            }
            {
                int[] numbersA = { 0, 2, 4, 5, 6, 8, 9 };
                int[] numbersB = { 1, 3, 5, 7, 8 };

                var uniqueNumbers = numbersA.Union( numbersB );

                Console.WriteLine( "Unique numbers from both arrays:" );
                foreach ( var n in uniqueNumbers )
                    Console.WriteLine( n );
            }
            {
                int[] numbersA = { 0, 2, 4, 5, 6, 8, 9 };
                int[] numbersB = { 1, 3, 5, 7, 8 };

                var commonNumbers = numbersA.Intersect( numbersB );

                Console.WriteLine( "Common numbers shared by both arrays:" );
                foreach ( var n in commonNumbers )
                    Console.WriteLine( n );
            }
            {
                int[] numbersA = { 0, 2, 4, 5, 6, 8, 9 };
                int[] numbersB = { 1, 3, 5, 7, 8 };

                IEnumerable<int> aOnlyNumbers = numbersA.Except( numbersB );

                Console.WriteLine( "Numbers in first array but not second array:" );
                foreach ( var n in aOnlyNumbers )
                    Console.WriteLine( n );
            }

            // Operatory konwersji
            {
                double[] doubles = { 1.7, 2.3, 1.9, 4.1, 2.9 };
                var sortedDoubles =
                    from d in doubles
                    orderby d descending
                    select d;
                var doublesArray = sortedDoubles.ToArray();

                Console.WriteLine( "Every other double from highest to lowest:" );
                for ( int d = 0; d < doublesArray.Length; d += 2 )
                    Console.WriteLine( doublesArray[d] );
            }
            {
                string[] words = { "cherry", "apple", "blueberry" };
                var sortedWords =
                    from w in words
                    orderby w
                    select w;
                var wordList = sortedWords.ToList();
                Console.WriteLine( "The sorted word list:" );
                foreach ( var w in wordList )
                {
                    Console.WriteLine( w );
                }
            }
            {
                var scoreRecords = new[] { new {Name = "Alice", Score = 50},
                                new {Name = "Bob" , Score = 40},
                                new {Name = "Cathy", Score = 45}
                              };
                var scoreRecordsDict = scoreRecords.ToDictionary( sr => sr.Name );
                Console.WriteLine( "Bob's score: {0}", scoreRecordsDict["Bob"] );
            }
            {
                object[] numbers = { null, 1.0, "two", 3, 4.0f, 5, "six", 7.0 };
                var doubles = numbers.OfType<double>();
                Console.WriteLine( "Numbers stored as doubles:" );
                foreach ( var d in doubles )
                {
                    Console.WriteLine( d );
                }
            }

            // Pobranie wybranego obiektu

            {
                List<Osoba> pusta = new List<Osoba>();
                var o1 = osoby.First();
                var o2 = osoby.First( o => o.Nazwisko.Contains( "03" ) );
                var o3 = osoby.FirstOrDefault( o => o.Nazwisko.Contains( "xx" ) );
                // var o4 = osoby.Single(); // to by zadziałało, gdyby w osoby był jeden obiekt
                var o5 = osoby.ElementAt( 3 );
                var o6 = osoby.ElementAtOrDefault( 30 );
                var o7 = pusta.DefaultIfEmpty( new Osoba { Imie = "Jan", Nazwisko = "Pusty" } );
            }

            // Operatory Any, All, Contains

            {
                var res1 = osoby.All( o => o.Nazwisko.StartsWith( "Kowal" ) );
                var res2 = osoby.Any( o => o.Nazwisko.EndsWith( "03" ) );
                var res3 = osoby.Contains( osoby[0] );
            }

            // Funkcje agregujące
            {
                int[] numbers = { 5, 4, 1, 3, 9, 8, 6, 7, 2, 0 };
                int oddNumbers = numbers.Count( n => n % 2 == 1 );
                Console.WriteLine( "There are {0} odd numbers in the list.", oddNumbers );
            }
            {
                string[] words = { "cherry", "apple", "blueberry" };
                double totalChars = words.Sum( w => w.Length );
                Console.WriteLine( "There are a total of {0} characters in these words.", totalChars );
            }
        }
    }
    public class AnagramEqualityComparer : IEqualityComparer<string>
    {
        public bool Equals( string x, string y )
        {
            return getCanonicalString( x ) == getCanonicalString( y );
        }

        public int GetHashCode( string obj )
        {
            return getCanonicalString( obj ).GetHashCode();
        }

        private string getCanonicalString( string word )
        {
            char[] wordChars = word.ToCharArray();
            Array.Sort<char>( wordChars );
            return new string( wordChars );
        }
    }
}
